from cg.apps.lims.orderform import CASE_PROJECT_TYPES, expand_case
from cg.constants import Pipeline
from cg.exc import OrderFormError
from cg.meta.orders.status import StatusHandler

OPTIONAL_KEYS = (
    "container_name",
    "quantity",
    "volume",
    "concentration",
    "status",
    "comment",
    "capture_kit",
    "mother",
    "father",
)


def get_project_type(samples: [dict]) -> str:
    """Determine the project type."""

    data_analyses = set(sample.get("data_analysis", "mip-dna").lower() for sample in samples)

    if len(data_analyses) != 1:
        raise OrderFormError(f"mixed 'Data Analysis' types: {', '.join(data_analyses)}")

    if data_analyses == {"mip-dna"}:
        project_type = "mip-dna"
    elif data_analyses == {"fluffy"}:
        project_type = "rml"
    elif data_analyses == {"balsamic"}:
        project_type = "balsamic"
    else:
        raise OrderFormError(f"Unsupported json orderform: {data_analyses}")

    return project_type


def parse_json(indata: dict) -> dict:
    """Parse JSON from LIMS export."""

    samples = indata.get("samples")

    if not samples:
        raise OrderFormError("orderform doesn't contain any samples")

    project_type = get_project_type(samples)
    customer_id = indata["customer"].lower()
    comment = indata.get("comment")

    if project_type in CASE_PROJECT_TYPES:
        parsed_cases = StatusHandler.group_cases(samples)
        items = []
        for case_id, parsed_case in parsed_cases.items():
            case_data = expand_case(case_id, parsed_case)
            items.append(case_data)
    else:
        items = samples

    data = {
        "customer": customer_id,
        "items": items,
        "project_type": project_type,
        "comment": comment,
    }

    return data


def expand_case(case_id: str, parsed_case: dict) -> dict:
    """Fill-in information about families."""
    new_case = {"name": case_id, "samples": []}
    samples = parsed_case

    require_qcoks = set(raw_sample["require_qcok"] for raw_sample in samples)
    new_case["require_qcok"] = True in require_qcoks

    priorities = set(raw_sample["priority"].lower() for raw_sample in samples)
    if len(priorities) != 1:
        raise OrderFormError(f"multiple values for 'Priority' for case: {case_id}")
    new_case["priority"] = priorities.pop()

    gene_panels = set()
    for raw_sample in samples:
        if raw_sample.get("panels"):
            gene_panels.update(raw_sample["panels"])

        new_sample = {}
        well_position_raw = raw_sample.get("well_position")
        if well_position_raw:
            new_sample["well_position"] = (
                ":".join(well_position_raw) if ":" not in well_position_raw else well_position_raw
            )

        for key, value in raw_sample.items():
            if key not in ["panels", "well_position"]:
                new_sample[key] = value

        new_case["samples"].append(new_sample)

    if gene_panels:
        new_case["panels"] = list(gene_panels)

    return new_case



import logging
import re
from typing import List

import click
from tabulate import tabulate

from cg.store import Store

LOG = logging.getLogger(__name__)
SAMPLE_HEADERS = ["Sample", "Name", "Customer", "Application", "State", "Priority", "External?"]
FAMILY_HEADERS = ["Family", "Name", "Customer", "Priority", "Panels", "Action"]
LINK_HEADERS = ["Sample", "Mother", "Father"]
FLOWCELL_HEADERS = ["Flowcell", "Type", "Sequencer", "Date", "Archived?", "Status"]


@click.group(invoke_without_command=True)
@click.option("-i", "identifier", help="made a guess what type you are looking for")
@click.pass_context
def get(context: click.Context, identifier: str):
    """Get information about records in the database."""
    context.obj["status_db"] = Store(context.obj["database"])
    if identifier and re.match(r"^[A-Z]{3}[0-9]{4,5}[A-Z]{1}[1-9]{1,3}$", identifier):
        context.invoke(sample, sample_ids=[identifier])
    elif identifier and re.match(r"^[a-z]*$", identifier):
        # try family information
        context.invoke(family, family_ids=[identifier])
    elif identifier and re.match(r"^[HC][A-Z0-9]{8}$", identifier):
        # try flowcell information
        context.invoke(flowcell, flowcell_id=identifier)
    elif identifier:
        LOG.error(f"{identifier}: can't predict identifier")
        context.abort()


@get.command()
@click.option("--families/--no-families", default=True, help="display related families")
@click.option("-f", "--flowcells", is_flag=True, help="display related flowcells")
@click.argument("sample_ids", nargs=-1)
@click.pass_context
def sample(context: click.Context, families: bool, flowcells: bool, sample_ids: List[str]):
    """Get information about a sample."""
    for sample_id in sample_ids:
        LOG.debug("%s: get info about sample", sample_id)
        sample_obj = context.obj["status_db"].sample(sample_id)
        if sample_obj is None:
            LOG.warning(f"{sample_id}: sample doesn't exist")
            continue
        row = [
            sample_obj.internal_id,
            sample_obj.name,
            sample_obj.customer.internal_id,
            sample_obj.application_version.application.tag,
            sample_obj.state,
            sample_obj.priority_human,
            "Yes" if sample_obj.application_version.application.is_external else "No",
        ]
        click.echo(tabulate([row], headers=SAMPLE_HEADERS, tablefmt="psql"))
        if families:
            family_ids = [link_obj.family.internal_id for link_obj in sample_obj.links]
            context.invoke(family, family_ids=family_ids, samples=False)
        if flowcells:
            for flowcell_obj in sample_obj.flowcells:
                LOG.debug(f"{flowcell_obj.name}: get info about flowcell")
                context.invoke(flowcell, flowcell_id=flowcell_obj.name, samples=False)


@get.command()
@click.argument("family_id")
@click.pass_context
def relations(context: click.Context, family_id: str):
    """Get information about a family relations."""

    family_obj = context.obj["status_db"].family(family_id)
    if family_obj is None:
        LOG.error("%s: family doesn't exist", family_id)
        context.abort()

    LOG.debug("%s: get info about family relations", family_obj.internal_id)

    for link_obj in family_obj.links:

        row = [
            link_obj.sample.internal_id if link_obj.sample else "",
            link_obj.mother.internal_id if link_obj.mother else "",
            link_obj.father.internal_id if link_obj.father else "",
        ]
        click.echo(tabulate([row], headers=LINK_HEADERS, tablefmt="psql"))


@get.command()
@click.option("-c", "--customer", help="internal id for customer to filter by")
@click.option("-n", "--name", is_flag=True, help="search family by name")
@click.option("--samples/--no-samples", default=True, help="display related samples")
@click.option("--relate/--no-relate", default=True, help="display relations to samples")
@click.argument("family_ids", nargs=-1)
@click.pass_context
def family(
    context: click.Context,
    customer: str,
    name: bool,
    samples: bool,
    relate: bool,
    family_ids: List[str],
):
    """Get information about a family."""
    if name:
        customer_obj = context.obj["status_db"].customer(customer)
        if customer_obj is None:
            LOG.error(f"{customer}: customer not found")
            context.abort()
        families = context.obj["status_db"].families(customer=customer_obj, enquiry=family_ids[-1])
    else:
        families = []
        for family_id in family_ids:
            family_obj = context.obj["status_db"].family(family_id)
            if family_obj is None:
                LOG.error(f"{family_id}: family doesn't exist")
                context.abort()
            families.append(family_obj)

    for family_obj in families:
        LOG.debug(f"{family_obj.internal_id}: get info about family")
        row = [
            family_obj.internal_id,
            family_obj.name,
            family_obj.customer.internal_id,
            family_obj.priority_human,
            ", ".join(family_obj.panels),
            family_obj.action or "NA",
        ]
        click.echo(tabulate([row], headers=FAMILY_HEADERS, tablefmt="psql"))
        if relate:
            context.invoke(relations, family_id=family_obj.internal_id)
        if samples:
            sample_ids = [link_obj.sample.internal_id for link_obj in family_obj.links]
            context.invoke(sample, sample_ids=sample_ids, families=False)


@get.command()
@click.option("--samples/--no-samples", default=True, help="display related samples")
@click.argument("flowcell_id")
@click.pass_context
def flowcell(context: click.Context, samples: bool, flowcell_id: str):
    """Get information about a flowcell and the samples on it."""
    flowcell_obj = context.obj["status_db"].flowcell(flowcell_id)
    if flowcell_obj is None:
        LOG.error(f"{flowcell_id}: flowcell not found")
        context.abort()
    row = [
        flowcell_obj.name,
        flowcell_obj.sequencer_type,
        flowcell_obj.sequencer_name,
        flowcell_obj.sequenced_at.date(),
        flowcell_obj.archived_at.date() if flowcell_obj.archived_at else "No",
        flowcell_obj.status,
    ]
    click.echo(tabulate([row], headers=FLOWCELL_HEADERS, tablefmt="psql"))
    if samples:
        sample_ids = [sample_obj.internal_id for sample_obj in flowcell_obj.samples]
        if sample_ids:
            context.invoke(sample, sample_ids=sample_ids, families=False)
        else:
            LOG.warning("no samples found on flowcell")



"""CLI support to start microsalt"""

import json
import logging
from pathlib import Path
from typing import Any

import click

from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Pipeline
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.store import Store

LOG = logging.getLogger(__name__)

OPTION_DRY_RUN = click.option(
    "-d",
    "--dry-run",
    help="Print command to console without executing",
    is_flag=True,
)
OPTION_SAMPLE = click.option(
    "-s",
    "--sample",
    help="Start workflow by providing a sample_id (By default case_id will be used)",
    is_flag=True,
)
OPTION_TICKET = click.option(
    "-t",
    "--ticket",
    help="Start workflow by providing a ticket_id (By default case_id will be used)",
    is_flag=True,
)
ARGUMENT_UNIQUE_IDENTIFIER = click.argument("unique_id", required=True, type=click.STRING)


@click.group(invoke_without_command=True)
@click.pass_context
def microsalt(context: click.Context):
    """Microbial workflow"""
    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None
    microsalt_analysis_api = MicrosaltAnalysisAPI(
        db=Store(context.obj["database"]),
        hk_api=HousekeeperAPI(context.obj),
        lims_api=LimsAPI(context.obj),
        hermes_api=HermesApi(context.obj),
        config=context.obj["microsalt"],
    )
    context.obj["microsalt_analysis_api"] = microsalt_analysis_api


@microsalt.command()
@OPTION_TICKET
@OPTION_SAMPLE
@ARGUMENT_UNIQUE_IDENTIFIER
@click.pass_context
def link(context: click.Context, ticket: bool, sample: bool, unique_id: str):
    """Link microbial FASTQ files for a SAMPLE_ID"""

    microsalt_analysis_api: MicrosaltAnalysisAPI = context.obj["microsalt_analysis_api"]

    case_id, sample_id = microsalt_analysis_api.resolve_case_sample_id(
        sample=sample, ticket=ticket, unique_id=unique_id
    )

    if not microsalt_analysis_api.check_flowcells_on_disk(case_id=case_id, sample_id=sample_id):
        raise click.Abort
    microsalt_analysis_api.link_samples(
        case_id=case_id,
        sample_id=sample_id,
    )


@microsalt.command("config-case")
@OPTION_DRY_RUN
@OPTION_TICKET
@OPTION_SAMPLE
@ARGUMENT_UNIQUE_IDENTIFIER
@click.pass_context
def config_case(context: click.Context, dry_run: bool, ticket: bool, sample: bool, unique_id: str):
    """ Create a config file on case level for microSALT """

    microsalt_analysis_api: MicrosaltAnalysisAPI = context.obj["microsalt_analysis_api"]

    case_id, sample_id = microsalt_analysis_api.resolve_case_sample_id(
        sample=sample, ticket=ticket, unique_id=unique_id
    )

    sample_objs = microsalt_analysis_api.get_samples(case_id=case_id, sample_id=sample_id)

    if not sample_objs:
        LOG.error("No sample found for that ticket/sample_id")
        raise click.Abort

    parameters = [microsalt_analysis_api.get_parameters(sample_obj) for sample_obj in sample_objs]

    filename = sample_id or case_id
    config_case_path: Path = microsalt_analysis_api.get_config_path(filename=filename)
    if dry_run:
        click.echo(json.dumps(parameters, indent=4, sort_keys=True))
        return

    with open(config_case_path, "w") as outfile:
        json.dump(parameters, outfile, indent=4, sort_keys=True)
    LOG.info("Saved config %s", config_case_path)


@microsalt.command()
@OPTION_DRY_RUN
@OPTION_TICKET
@OPTION_SAMPLE
@ARGUMENT_UNIQUE_IDENTIFIER
@click.option(
    "-c",
    "--config-case",
    "config_case_path",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="optionally change the config-case",
)
@click.pass_context
def run(
    context: click.Context,
    dry_run: bool,
    config_case_path: click.Path,
    ticket: bool,
    sample: bool,
    unique_id: Any,
):
    """ Start microSALT with an order_id """

    microsalt_analysis_api: MicrosaltAnalysisAPI = context.obj["microsalt_analysis_api"]

    case_id, sample_id = microsalt_analysis_api.resolve_case_sample_id(
        sample=sample, ticket=ticket, unique_id=unique_id
    )

    fastq_path: Path = microsalt_analysis_api.get_case_fastq_path(case_id=case_id)

    if not config_case_path:
        filename = sample_id or case_id
        config_case_path: Path = microsalt_analysis_api.get_config_path(filename=filename)

    if not sample_id:
        analyse_command = [
            "analyse",
            config_case_path.absolute().as_posix(),
            "--input",
            fastq_path.absolute().as_posix(),
        ]
    else:
        analyse_command = [
            "analyse",
            config_case_path.absolute().as_posix(),
            "--input",
            Path(fastq_path, sample_id).absolute().as_posix(),
        ]
    microsalt_analysis_api.process.run_command(parameters=analyse_command, dry_run=dry_run)

    if not sample_id and not dry_run:
        microsalt_analysis_api.set_statusdb_action(case_id=case_id, action="running")


@microsalt.command()
@ARGUMENT_UNIQUE_IDENTIFIER
@click.pass_context
def store(context: click.Context, unique_id: str):
    microsalt_analysis_api: MicrosaltAnalysisAPI = context.obj["microsalt_analysis_api"]

    case_obj = microsalt_analysis_api.db.family(unique_id)
    if not case_obj:
        LOG.error("Please provide a valid case id!")
        raise click.Abort
    try:
        microsalt_analysis_api.store_microbial_analysis_housekeeper(case_id=unique_id)
        microsalt_analysis_api.store_microbial_analysis_statusdb(case_id=unique_id)
        microsalt_analysis_api.set_statusdb_action(case_id=unique_id, action=None)
    except Exception as error:
        microsalt_analysis_api.db.rollback()
        microsalt_analysis_api.hk.rollback()
        LOG.error(
            "Error storing deliverables for case %s - %s", unique_id, error.__class__.__name__
        )
        raise


@microsalt.command()
@ARGUMENT_UNIQUE_IDENTIFIER
@OPTION_DRY_RUN
@OPTION_TICKET
@OPTION_SAMPLE
@click.pass_context
def start(context: click.Context, ticket: bool, sample: bool, unique_id: str, dry_run: bool):
    LOG.info("Starting Microsalt workflow for %s", unique_id)
    context.invoke(link, ticket=ticket, sample=sample, unique_id=unique_id)
    context.invoke(config_case, ticket=ticket, sample=sample, unique_id=unique_id, dry_run=dry_run)
    context.invoke(run, ticket=ticket, sample=sample, unique_id=unique_id, dry_run=dry_run)


@microsalt.command("store-available")
@OPTION_DRY_RUN
@click.pass_context
def store_available(context: click.Context, dry_run: bool):
    microsalt_analysis_api: MicrosaltAnalysisAPI = context.obj["microsalt_analysis_api"]
    exit_code: int = EXIT_SUCCESS

    for case_obj in microsalt_analysis_api.get_deliverables_to_store():
        LOG.info("Storing deliverables for %s", case_obj.internal_id)
        if dry_run:
            continue
        try:
            context.invoke(store, unique_id=case_obj.internal_id)
        except Exception:
            exit_code = EXIT_FAIL

    if exit_code:
        raise click.Abort


@microsalt.command("start-available")
@OPTION_DRY_RUN
@click.pass_context
def start_available(context: click.Context, dry_run: bool):
    microsalt_analysis_api: MicrosaltAnalysisAPI = context.obj["microsalt_analysis_api"]
    exit_code: int = EXIT_SUCCESS
    for case_obj in microsalt_analysis_api.db.cases_to_analyze(pipeline=Pipeline.MICROSALT):
        if dry_run:
            LOG.info("Would have started workflow for case %s", case_obj.internal_id)
            continue
        try:
            context.invoke(start, unique_id=case_obj.internal_id)
        except Exception as error:
            LOG.error(
                "Error when starting analysis for case %s - %s",
                case_obj.internal_id,
                error.__class__.__name__,
            )
            exit_code = EXIT_FAIL

    if exit_code:
        raise click.Abort
